# test == !=
assert True == True
assert True != False
assert False == False
assert False != True

# test and/or/not
assert True and True
assert not (True and False)
assert True or True
assert True or False
assert not False
assert not (not True)

assert bool() == False
assert bool(0) == False
assert bool(1) == True
assert bool([]) == False
assert bool("abc") == True
assert bool([1,2]) == True
assert bool('') == False

# is operator
assert None == None
assert None is None
assert ... == ...
assert ... is ...
assert NotImplemented == NotImplemented
assert NotImplemented is NotImplemented

assert True is True
assert False is False

assert True + 1 == 2
assert True - 1 == 0
assert True * 3 == 3
assert 1 + True == 2
assert 1 - True == 0
assert 3 * True == 3
