#pragma once
// generated by prebuild.py

const char* load_kPythonLib(const char* name);

extern const char kPythonLibs_bisect[];
extern const char kPythonLibs_builtins[];
extern const char kPythonLibs_cmath[];
extern const char kPythonLibs_collections[];
extern const char kPythonLibs_dataclasses[];
extern const char kPythonLibs_datetime[];
extern const char kPythonLibs_functools[];
extern const char kPythonLibs_heapq[];
extern const char kPythonLibs_linalg[];
extern const char kPythonLibs_operator[];
extern const char kPythonLibs_typing[];
