def Linear(t: float) -> float: ...
def InSine(t: float) -> float: ...
def OutSine(t: float) -> float: ...
def InOutSine(t: float) -> float: ...
def InQuad(t: float) -> float: ...
def OutQuad(t: float) -> float: ...
def InOutQuad(t: float) -> float: ...
def InCubic(t: float) -> float: ...
def OutCubic(t: float) -> float: ...
def InOutCubic(t: float) -> float: ...
def InQuart(t: float) -> float: ...
def OutQuart(t: float) -> float: ...
def InOutQuart(t: float) -> float: ...
def InQuint(t: float) -> float: ...
def OutQuint(t: float) -> float: ...
def InOutQuint(t: float) -> float: ...
def InExpo(t: float) -> float: ...
def OutExpo(t: float) -> float: ...
def InOutExpo(t: float) -> float: ...
def InCirc(t: float) -> float: ...
def OutCirc(t: float) -> float: ...
def InOutCirc(t: float) -> float: ...
def InBack(t: float) -> float: ...
def OutBack(t: float) -> float: ...
def InOutBack(t: float) -> float: ...
def InElastic(t: float) -> float: ...
def OutElastic(t: float) -> float: ...
def InOutElastic(t: float) -> float: ...
def InBounce(t: float) -> float: ...
def OutBounce(t: float) -> float: ...
def InOutBounce(t: float) -> float: ...
