from vmath import vec3

def linear_srgb_to_srgb(rgb: vec3) -> vec3: ...
def srgb_to_linear_srgb(rgb: vec3) -> vec3: ...
def srgb_to_hsv(rgb: vec3) -> vec3: ...
def hsv_to_srgb(hsv: vec3) -> vec3: ...
def oklch_to_linear_srgb(lch: vec3) -> vec3: ...
def linear_srgb_to_oklch(rgb: vec3) -> vec3: ...
